import logging

from yupi import DiffMethod, Trajectory, WindowType

from pactus.dataset import Dataset, GeoLifeDataset, LangevinDataset, MnistStrokeDataset
from pactus.features import Featurizer
from pactus.models import (
    DecisionTreeModel,
    Evaluation,
    EvaluationComparison,
    KNeighborsModel,
    Model,
    RandomForestModel,
    SVMModel,
    TransformerModel,
)

__version__ = "0.1.0"

Trajectory.global_diff_method(
    method=DiffMethod.LINEAR_DIFF, window_type=WindowType.FORWARD
)

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s [%(levelname)s]: %(message)s"
)



"""
General configurations
"""
import os
from pathlib import Path

# -----------------------------------------------------------------------------
# Download configs
# -----------------------------------------------------------------------------
DOWNLOAD_CHUNCK_SIZE = 4096
PROGRESS_BAR_LENGTH = 50

# -----------------------------------------------------------------------------
# Dataset configs
# -----------------------------------------------------------------------------
#
# Dataset version: (doens't need to match the project version)
#   [major].[minor].[patch]
#      │       │       └── Change when minors changes are made in the code
#      │       │           related to the datasets. (no need for redownload
#      │       │           or regeneration of yupi trajectories)
#      │       │
#      │       └── Change when the yupi data changes (needs to be regenerated)
#      │
#      └── Change when the original dataset changes (needs redownload)
#
# Dataset structure: (mainly for downloadable datasets)
#  src
#  └── datasets
#      └── [dataset_name]
#          ├── raw_data
#          │   └── [unzipped dataset files]
#          ├── metadata.json
#          └── yupi_metadata.json
#
CACHE_PATH = os.environ.get("YUCA_CACHE_PATH", str(Path(__file__).parent / ".cache"))
DS_BASE_DIR = CACHE_PATH + "/datasets"
DS_DIR = DS_BASE_DIR + "/{0}"
DS_METADATA_FILE = DS_DIR + "/metadata.json"
DS_RAW_DIR = DS_DIR + "/raw_data"
DS_FEATS_DIR = DS_DIR + "/features"
DS_EVALS_DIR = DS_DIR + "/evaluations"



import logging
from datetime import datetime
from pathlib import Path
from typing import List, NamedTuple, Tuple

import numpy as np
from yupi import Trajectory

from pactus.dataset import Dataset, download_dataset
from pactus.dataset._utils import _get_progress_log

# Dataset metadata
NAME = "geolife"
VERSION = "0.1.0"  # See version description in config.py
DOWNLOAD_URL = (
    "https://download.microsoft.com/download/F/4/8/"
    "F4894AA5-FDBC-481E-9285-D5F8C4C4F039/Geolife%20Trajectories%201.3.zip"
)


LabelData = NamedTuple(
    "LabelData", [("start_dt", datetime), ("end_dt", datetime), ("name", str)]
)


Register = Tuple[float, float, datetime]
"""Register data (GPS point and time) in the form: lat, lon, time"""


class GeoLifeDataset(Dataset):
    """Class for the GeoLife dataset."""

    def __init__(self, redownload: bool = False, reyupify: bool = False):
        super().__init__(NAME, VERSION, redownload, reyupify)

    def fetch(self) -> None:
        download_dataset(DOWNLOAD_URL, self.name)

    def yupify(self):
        # Loads the raw data and preprocess it
        raw_metadata = []
        logging.info("Preprocessing GeoLife raw data")
        usr_folders_path = self.raw_dir / "Geolife Trajectories 1.3/Data"
        usr_folders = list(sorted(usr_folders_path.iterdir()))
        for i, usr in enumerate(usr_folders):
            print(_get_progress_log(i + 1, len(usr_folders)), end="\r")
            self._process_usr_trajs(usr, raw_metadata)

        # Load the preprocessed data and create the yupi trajectories
        logging.info("Creating yupi trajectories")
        trajs, lables = [], []
        for i, traj_data in enumerate(raw_metadata):
            print(_get_progress_log(i + 1, len(raw_metadata)), end="\r")
            x_data = traj_data["traj_data"][:, 0]
            y_data = traj_data["traj_data"][:, 1]
            t_data = traj_data["traj_data"][:, 2]
            trajs.append(Trajectory(x=x_data, y=y_data, t=t_data))
            lables.append(traj_data["label"])

        return trajs, lables

    def _load_labels(self, labels_file: Path) -> List[LabelData]:
        """Loads the labels of a user."""
        with open(labels_file, "r", encoding="utf-8") as l_file:
            labels = [self._parse_label(line) for line in l_file.readlines()[1:]]
        return labels

    def _process_usr_trajs(self, usr_folder: Path, raw_metadata: List[dict]) -> None:
        """Processes the trajectories of a user."""
        labels_file = usr_folder / "labels.txt"
        if not labels_file.exists():
            return

        labels = self._load_labels(labels_file)
        all_regs = self._load_registers(usr_folder)

        label_idx = 0
        regs_idx = 0
        label = labels[label_idx]
        traj: List[List[float]] = []
        while regs_idx < len(all_regs):
            lat, long, reg_dt = all_regs[regs_idx]

            # Register inside label bounds
            if label.start_dt <= reg_dt <= label.end_dt:
                time = (reg_dt - label.start_dt).seconds
                if not traj or time != traj[-1][-1]:
                    traj.append([lat, long, time])

            # Register outside label bounds
            elif label.start_dt < reg_dt:
                # Save the trajectory if it has at least 5 points
                if len(traj) > 5:
                    raw_metadata.append(
                        {
                            "id": f"{usr_folder.name}_{label_idx}",
                            "traj_data": np.array(traj),
                            "label": label.name,
                        }
                    )
                    traj.clear()

                # Move to the next label
                label_idx += 1
                if label_idx >= len(labels):
                    break
                label = labels[label_idx]
                continue  # Don't increment regs_idx yet
            regs_idx += 1

    def _load_registers(self, usr_folder: Path) -> List[Register]:
        """Loads all the registers of a user."""
        all_regs = []
        trajs_folder = usr_folder / "Trajectory"
        sorted_paths = sorted(trajs_folder.iterdir())
        for plt in sorted_paths:
            with open(plt, "r", encoding="utf-8") as reg:
                registers = [self._parse_register(line) for line in reg.readlines()[6:]]
                all_regs += registers
        return all_regs

    def _parse_register(self, line: str) -> Register:
        """Parses a register line."""
        reg = line.strip().split(",")
        lat, lon = float(reg[0]), float(reg[1])
        _dt = datetime.strptime(f"{reg[5]} {reg[6]}", "%Y-%m-%d %H:%M:%S")
        return lat, lon, _dt

    def _parse_label(self, line: str) -> LabelData:
        """Parses a label line."""
        items = line.split()
        start_dt = datetime.strptime(f"{items[0]} {items[1]}", "%Y/%m/%d %H:%M:%S")
        end_dt = datetime.strptime(f"{items[2]} {items[3]}", "%Y/%m/%d %H:%M:%S")
        clsf = items[4]
        return LabelData(start_dt, end_dt, clsf)



from pactus.dataset._dataset import Data, Dataset
from pactus.dataset._utils import download_dataset
from pactus.dataset.geolife_dataset import GeoLifeDataset
from pactus.dataset.langevin_dataset import LangevinDataset
from pactus.dataset.mnist_stroke_dataset import MnistStrokeDataset

__all__ = [
    "Data",
    "Dataset",
    "download_dataset",
    "GeoLifeDataset",
    "LangevinDataset",
    "MnistStrokeDataset",
]



from random import Random
from typing import Any, List, Tuple

from yupi import Trajectory
from yupi.generators import LangevinGenerator

from pactus.dataset import Dataset

# Dataset metadata
NAME = "langevin"
VERSION = "0.1.0"  # See version description in config.py


class LangevinDataset(Dataset):
    """Dataset of trajectories generated from the Langevin model."""

    def __init__(self, reyupify: bool = False):
        super().__init__(NAME, VERSION, False, reyupify)

    def yupify(self) -> Tuple[List[Trajectory], List[Any]]:
        trajs, labels = [], []
        seed = 0

        def get_seed() -> int:
            nonlocal seed
            seed += 1
            return seed

        rng = Random(0)

        total_time_mean = 10
        dt_mean = 0.1
        labels_config = {
            "type_1": {
                "gamma": 0.1,
                "sigma": 0.1,
                "count": 1000,
            },
            "type_2": {
                "gamma": 0.2,
                "sigma": 0.1,
                "count": 1000,
            },
        }

        for label, config in labels_config.items():
            assert isinstance(config["count"], int) and config["count"] > 0
            for _ in range(config["count"]):
                gamma, sigma = config["gamma"], config["sigma"]
                lg_gen = LangevinGenerator(
                    dim=2,
                    N=1,
                    dt=abs(rng.normalvariate(dt_mean, 0.1)),
                    T=abs(rng.normalvariate(total_time_mean, 2)),
                    seed=get_seed(),
                    gamma=gamma,
                    sigma=sigma,
                )
                trajs.append(lg_gen.generate()[0])
                labels.append(label)

        return trajs, labels



import logging
from pathlib import Path
from string import Template
from typing import List, Tuple

import numpy as np
from yupi import Trajectory

from pactus.dataset import Dataset, download_dataset

# Dataset metadata
NAME = "mnist_stroke"
VERSION = "0.1.0"  # See version description in config.py
DOWNLOAD_URL = (
    "https://github.com/edwin-de-jong/mnist-digits-stroke-sequence-data/"
    "raw/master/sequences.tar.gz"
)
TRAIN_LABELS_URL = "http://yann.lecun.com/exdb/mnist/train-labels-idx1-ubyte.gz"
TEST_LABELS_URL = "http://yann.lecun.com/exdb/mnist/t10k-labels-idx1-ubyte.gz"


class MnistStrokeDataset(Dataset):
    """Class for the MNIST stroke sequence dataset."""

    def __init__(self, redownload: bool = False, reyupify: bool = False):
        super().__init__(NAME, VERSION, redownload, reyupify)

    def fetch(self) -> None:
        download_dataset(DOWNLOAD_URL, self.name)
        download_dataset(TRAIN_LABELS_URL, self.name)
        download_dataset(TEST_LABELS_URL, self.name)

    def _read_labels(self, label_file: Path, count: int) -> List[str]:
        with open(label_file, "rb") as f:
            f.read(8)  # discard header info
            return [str(l) for l in f.read(count)]

    def _read_traj(self, trajectory_file: Path) -> Trajectory:
        # Load raw data from the stroke
        traj = np.loadtxt(trajectory_file, delimiter=",", skiprows=1)

        # Filter tokens from changes
        points = [(i, j) for i, j in traj if i >= 0 and j >= 0]

        return Trajectory(points=points)

    def _read_trajs(
        self, sequence_folder: Path, template: Template, count: int
    ) -> List[Trajectory]:
        trajs = []
        for i in range(count):
            traj_path = sequence_folder / template.substitute(id=i)
            trajs.append(self._read_traj(traj_path))

        return trajs

    def _yupify_mnist(
        self, sequence_folder: Path, label_file: Path, template: Template, count: int
    ) -> Tuple[List[Trajectory], List[str]]:
        """Yupifies a part of the dataset"""

        labels = self._read_labels(label_file, count)
        trajs = self._read_trajs(sequence_folder, template, count)

        return trajs, labels

    def yupify(self) -> Tuple[List[Trajectory], List[str]]:
        # Loads the raw data and preprocess it
        logging.info("Preprocessing MNIST stroke raw data")
        sequence_path = self.raw_dir / "sequences"
        train_labels_path = self.raw_dir / "train-labels-idx1-ubyte"
        test_labels_path = self.raw_dir / "t10k-labels-idx1-ubyte"

        logging.info("Yupifying train dataset...")
        train_template = Template("trainimg-$id-points.txt")
        trajs_train, labels_train = self._yupify_mnist(
            sequence_path, train_labels_path, template=train_template, count=60000
        )

        logging.info("Yupifying test dataset...")
        test_template = Template("testimg-$id-points.txt")
        trajs_test, labels_test = self._yupify_mnist(
            sequence_path, test_labels_path, template=test_template, count=10000
        )

        return trajs_train + trajs_test, labels_train + labels_test



import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import layers


def transformer_encoder(inputs, head_size, num_heads, ff_dim, dropout=0.0, mask=None):
    if mask is not None:
        mask = mask[:, tf.newaxis, tf.newaxis, :]

    # Normalization and Attention
    x = layers.LayerNormalization(epsilon=1e-6)(inputs)
    x = layers.MultiHeadAttention(
        key_dim=head_size, num_heads=num_heads, dropout=dropout
    )(x, x, attention_mask=mask)
    x = layers.Dropout(dropout)(x)
    res = x + inputs

    # Feed Forward Part
    x = layers.LayerNormalization(epsilon=1e-6)(res)
    x = layers.Conv1D(filters=ff_dim, kernel_size=1, activation="relu")(x)
    x = layers.Dropout(dropout)(x)
    x = layers.Conv1D(filters=inputs.shape[-1], kernel_size=1)(x)
    return x + res


def build_model(
    n_classes,
    input_shape,
    head_size,
    num_heads,
    ff_dim,
    num_transformer_blocks,
    mlp_units,
    dropout=0.0,
    mlp_dropout=0.0,
    input_mask=None,
) -> keras.Model:
    inputs = keras.Input(shape=input_shape)
    x = inputs
    for _ in range(num_transformer_blocks):
        x = transformer_encoder(x, head_size, num_heads, ff_dim, dropout, input_mask)

    x = layers.GlobalAveragePooling2D(data_format="channels_first")(x)
    for dim in mlp_units:
        x = layers.Dense(dim, activation="relu")(x)
        x = layers.Dropout(mlp_dropout)(x)
    outputs = layers.Dense(n_classes, activation="softmax")(x)
    return keras.Model(inputs, outputs)



from pactus.models.decision_tree_model import DecisionTreeModel
from pactus.models.evaluation import Evaluation
from pactus.models.evaluation_comparison import EvaluationComparison
from pactus.models.kneighbors_model import KNeighborsModel
from pactus.models.model import Model
from pactus.models.random_forest_model import RandomForestModel
from pactus.models.svm_model import SVMModel
from pactus.models.transformer_model import TransformerModel



"""
This is not intended to be used directly, but as a template for creating new
datasets.
"""

from typing import Any, List, Tuple

from yupi import Trajectory

from pactus.dataset import Dataset

# Dataset metadata
NAME = "template"  # Name of the dataset
VERSION = "0.1.0"  # See version description in config.py
# Sometimes the dataset needs to be downloaded from a url. If that is the case,
# this is also a good place to put the url.
URL = "https://example.com/dataset.zip"

# The name of the class should be the same as the name of the dataset, followed
# by "Dataset".
class TemplateDataset(Dataset):
    """Template dataset."""

    # Call the super class constructor with dataset info
    # The refetch and reyupify arguments are used to force the dataset to be
    # redownloaded or reconverted to yupi trajectories.
    def __init__(self, refetch: bool = False, reyupify: bool = False):
        super().__init__(NAME, VERSION, refetch, reyupify)

    # 'fetch' is the method you need to override to download the dataset
    # if needed. The self.raw_dir variable is the path to the folder
    # where the dataset should be stored.
    #
    # If the dataset is generated, this method can be left empty.
    # Use the 'yupify' method instead.
    def fetch(self) -> None:
        """Downloads the dataset in case needed"""

    # 'yupify' is the method you need to override to parse or generate the
    # dataset and convert it to yupi trajectories. The method should return
    # a tuple containing a list of yupi trajectories and a list of labels.
    #
    # The labels can be anything, but they must be the same length as the
    # list of trajectories and the values must be storable in a json file.
    # (e.g., strings or numbers usually)
    def yupify(self) -> Tuple[List[Trajectory], List[Any]]:
        """Parses or generates the dataset and convert it to yupi trajectories"""
        raise NotImplementedError



from collections import Counter
from functools import reduce
from operator import or_
from typing import Callable, Union

import numpy as np
from yupi.stats._stats import _kurtosis
from yupi.trajectory import Trajectory, Vector

FEAT_DICT = {}
__FEAT_VAL = 1


def __new_feat(feat_name: Union[str, None] = None) -> int:
    global __FEAT_VAL  # pylint: disable=global-statement
    if feat_name is not None:
        FEAT_DICT[__FEAT_VAL] = feat_name
    feat_val = __FEAT_VAL
    __FEAT_VAL <<= 1
    return feat_val


def __all_feat_val() -> int:
    return __new_feat() - 1


DISTANCE = __new_feat("DISTANCE")

# Velocity features
VEL_MEAN = __new_feat("VEL_MEAN")
VEL_MEDIAN = __new_feat("VEL_MEDIAN")
VEL_MODE = __new_feat("VEL_MODE")
VEL_KURTOSIS = __new_feat("VEL_KURTOSIS")
VEL_AUTO_CC = __new_feat("VEL_AUTO_CC")
VEL_MIN = __new_feat("VEL_MIN")
VEL_MAX = __new_feat("VEL_MAX")
VEL_RANGE = __new_feat("VEL_RANGE")
VEL_STD = __new_feat("VEL_STD")
VEL_VAR = __new_feat("VEL_VAR")
VEL_COEF_VAR = __new_feat("VEL_COEF_VAR")
VEL_IQR = __new_feat("VEL_IQR")
STOP_RATE = __new_feat("STOP_RATE")
VEL_CHANGE_RATE = __new_feat("VEL_CHANGE_RATE")

# Acceleration features
ACC_MEAN = __new_feat("ACC_MEAN")
ACC_MEDIAN = __new_feat("ACC_MEDIAN")
ACC_MODE = __new_feat("ACC_MODE")
ACC_KURTOSIS = __new_feat("ACC_KURTOSIS")
ACC_AUTO_CC = __new_feat("ACC_AUTO_CC")
ACC_MIN = __new_feat("ACC_MIN")
ACC_MAX = __new_feat("ACC_MAX")
ACC_RANGE = __new_feat("ACC_RANGE")
ACC_STD = __new_feat("ACC_STD")
ACC_VAR = __new_feat("ACC_VAR")
ACC_COEF_VAR = __new_feat("ACC_COEF_VAR")
ACC_IQR = __new_feat("ACC_IQR")

# Acceleration change rate features
ACC_CHANGE_RATE_MEAN = __new_feat("ACC_CHANGE_RATE_MEAN")
ACC_CHANGE_RATE_MEDIAN = __new_feat("ACC_CHANGE_RATE_MEDIAN")
ACC_CHANGE_RATE_MODE = __new_feat("ACC_CHANGE_RATE_MODE")
ACC_CHANGE_RATE_KURTOSIS = __new_feat("ACC_CHANGE_RATE_KURTOSIS")
ACC_CHANGE_RATE_AUTO_CC = __new_feat("ACC_CHANGE_RATE_AUTO_CC")
ACC_CHANGE_RATE_MIN = __new_feat("ACC_CHANGE_RATE_MIN")
ACC_CHANGE_RATE_MAX = __new_feat("ACC_CHANGE_RATE_MAX")
ACC_CHANGE_RATE_RANGE = __new_feat("ACC_CHANGE_RATE_RANGE")
ACC_CHANGE_RATE_STD = __new_feat("ACC_CHANGE_RATE_STD")
ACC_CHANGE_RATE_VAR = __new_feat("ACC_CHANGE_RATE_VAR")
ACC_CHANGE_RATE_COEF_VAR = __new_feat("ACC_CHANGE_RATE_COEF_VAR")
ACC_CHANGE_RATE_IQR = __new_feat("ACC_CHANGE_RATE_IQR")

# Angle features
ANGLE_MEAN = __new_feat("ANGLE_MEAN")
ANGLE_MEDIAN = __new_feat("ANGLE_MEDIAN")
ANGLE_MODE = __new_feat("ANGLE_MODE")
ANGLE_KURTOSIS = __new_feat("ANGLE_KURTOSIS")
ANGLE_AUTO_CC = __new_feat("ANGLE_AUTO_CC")
ANGLE_MIN = __new_feat("ANGLE_MIN")
ANGLE_MAX = __new_feat("ANGLE_MAX")
ANGLE_RANGE = __new_feat("ANGLE_RANGE")
ANGLE_STD = __new_feat("ANGLE_STD")
ANGLE_VAR = __new_feat("ANGLE_VAR")
ANGLE_COEF_VAR = __new_feat("ANGLE_COEF_VAR")
ANGLE_IQR = __new_feat("ANGLE_IQR")

# Turning angle features
TURNING_ANGLE_MEAN = __new_feat("TURNING_ANGLE_MEAN")
TURNING_ANGLE_MEDIAN = __new_feat("TURNING_ANGLE_MEDIAN")
TURNING_ANGLE_MODE = __new_feat("TURNING_ANGLE_MODE")
TURNING_ANGLE_KURTOSIS = __new_feat("TURNING_ANGLE_KURTOSIS")
TURNING_ANGLE_AUTO_CC = __new_feat("TURNING_ANGLE_AUTO_CC")
TURNING_ANGLE_MIN = __new_feat("TURNING_ANGLE_MIN")
TURNING_ANGLE_MAX = __new_feat("TURNING_ANGLE_MAX")
TURNING_ANGLE_RANGE = __new_feat("TURNING_ANGLE_RANGE")
TURNING_ANGLE_STD = __new_feat("TURNING_ANGLE_STD")
TURNING_ANGLE_VAR = __new_feat("TURNING_ANGLE_VAR")
TURNING_ANGLE_COEF_VAR = __new_feat("TURNING_ANGLE_COEF_VAR")
TURNING_ANGLE_IQR = __new_feat("TURNING_ANGLE_IQR")

# Turning angle change rate features
TURNING_ANGLE_CHANGE_RATE_MEAN = __new_feat("TURNING_ANGLE_CHANGE_RATE_MEAN")
TURNING_ANGLE_CHANGE_RATE_MEDIAN = __new_feat("TURNING_ANGLE_CHANGE_RATE_MEDIAN")
TURNING_ANGLE_CHANGE_RATE_MODE = __new_feat("TURNING_ANGLE_CHANGE_RATE_MODE")
TURNING_ANGLE_CHANGE_RATE_KURTOSIS = __new_feat("TURNING_ANGLE_CHANGE_RATE_KURTOSIS")
TURNING_ANGLE_CHANGE_RATE_AUTO_CC = __new_feat("TURNING_ANGLE_CHANGE_RATE_AUTO_CC")
TURNING_ANGLE_CHANGE_RATE_MIN = __new_feat("TURNING_ANGLE_CHANGE_RATE_MIN")
TURNING_ANGLE_CHANGE_RATE_MAX = __new_feat("TURNING_ANGLE_CHANGE_RATE_MAX")
TURNING_ANGLE_CHANGE_RATE_RANGE = __new_feat("TURNING_ANGLE_CHANGE_RATE_RANGE")
TURNING_ANGLE_CHANGE_RATE_STD = __new_feat("TURNING_ANGLE_CHANGE_RATE_STD")
TURNING_ANGLE_CHANGE_RATE_VAR = __new_feat("TURNING_ANGLE_CHANGE_RATE_VAR")
TURNING_ANGLE_CHANGE_RATE_COEF_VAR = __new_feat("TURNING_ANGLE_CHANGE_RATE_COEF_VAR")
TURNING_ANGLE_CHANGE_RATE_IQR = __new_feat("TURNING_ANGLE_CHANGE_RATE_IQR")

ALL = __all_feat_val()


def group(key: Callable[[str], bool]) -> int:
    """Grups features according a predicate to their names"""
    return reduce(or_, [feat for feat, name in FEAT_DICT.items() if key(name)])


# Gruped by feat type
VEL_FEATS = group(lambda n: n.startswith("VEL"))
ACC_CHANGE_RATE_FEATS = group(lambda n: n.startswith("ACC_CHANGE_RATE"))
ACC_FEATS = group(lambda n: n.startswith("ACC") and not n.startswith("ACC_CHANGE_RATE"))
ANGLE_FEATS = group(lambda n: n.startswith("ANGLE"))
TURNING_ANGLE_CHANGE_RATE_FEATS = group(
    lambda n: n.startswith("TURNING_ANGLE_CHANGE_RATE")
)
TURNING_ANGLE_FEATS = group(
    lambda n: n.startswith("TURNING_ANGLE")
    and not n.startswith("TURNING_ANGLE_CHANGE_RATE")
)

# Gruped by stat type
MEAN_FEATS = group(lambda n: n.endswith("MEAN"))
MEDIAN_FEATS = group(lambda n: n.endswith("MEDIAN"))
MODE_FEATS = group(lambda n: n.endswith("MODE"))
KURTOSIS_FEATS = group(lambda n: n.endswith("KURTOSIS"))
AUTO_CC_FEATS = group(lambda n: n.endswith("AUTO_CC"))
MIN_FEATS = group(lambda n: n.endswith("MIN"))
MAX_FEATS = group(lambda n: n.endswith("MAX"))
RANGE_FEATS = group(lambda n: n.endswith("RANGE"))
STD_FEATS = group(lambda n: n.endswith("STD"))
COEF_VAR_FEATS = group(lambda n: n.endswith("COEF_VAR"))
VAR_FEATS = group(lambda n: n.endswith("VAR") and not n.endswith("COEF_VAR"))
IQR_FEATS = group(lambda n: n.endswith("IQR"))

FEAT_NAMES = list(FEAT_DICT.values())
FEAT_VALUES = list(FEAT_DICT.keys())


def _acc_k(values: Vector, k: int, mean: float) -> float:
    N = len(values)  # pylint: disable=invalid-name
    _sum = np.sum((values[1 : N - k] - mean) * (values[1 + k : N] - mean))
    return _sum / N


def get_feat(traj: Trajectory, feat: int, **kwargs) -> float:
    """Get feature value from trajectory.

    Parameters
    ----------
    traj : Trajectory
        Trajectory.
    feat : int
        Feature.

    Returns
    -------
    float
        Feature value.

    """

    if feat == DISTANCE:
        norm = traj.r.delta.norm
        assert isinstance(norm, Vector)
        return sum(norm)

    if feat == STOP_RATE:
        vel = traj.v.delta.norm
        assert "stop_rate_threshold" in kwargs
        threshold = kwargs["stop_rate_threshold"]
        return np.sum(vel < threshold) / get_feat(traj, DISTANCE)

    if feat == VEL_CHANGE_RATE:
        vel = traj.v.delta.norm
        assert isinstance(vel, Vector)
        subs = np.abs(np.diff(vel))
        vel[vel == 0] = np.inf
        v_rate = subs / vel[:-1]
        assert "vel_change_rate_threshold" in kwargs
        threshold = kwargs["vel_change_rate_threshold"]
        return np.sum(v_rate > threshold) / get_feat(traj, DISTANCE)

    values = None

    if feat & VEL_FEATS:
        values = traj.v.delta.norm
        assert isinstance(values, Vector)
    elif feat & ACC_FEATS:
        values = traj.a.delta.norm
        assert isinstance(values, Vector)
    elif feat & ACC_CHANGE_RATE_FEATS:
        acc = traj.a.delta.norm
        assert isinstance(acc, Vector)
        delta_t = traj.t.delta
        values = np.diff(acc) / delta_t[1:]
    elif feat & ANGLE_FEATS:
        values = traj.turning_angles(accumulate=True)
    elif feat & TURNING_ANGLE_FEATS:
        values = traj.turning_angles(accumulate=False)
    elif feat & TURNING_ANGLE_CHANGE_RATE_FEATS:
        angle = traj.turning_angles(accumulate=False)
        assert isinstance(angle, Vector)
        delta_t = traj.t.delta
        values = np.diff(angle) / delta_t[2:]

    assert values is not None, "Feature not found"
    assert (
        isinstance(values, Vector) and values.shape[0] > 0
    ), f"Empty values for feature {FEAT_DICT[feat]}"

    assert (
        len(values.shape) == 1
    ), f"Only 1D vectors are supported. Feature: {FEAT_DICT[feat]}"
    assert np.nan not in values, f"NaN values for feature {FEAT_DICT[feat]}"

    if feat & MEAN_FEATS:
        return float(np.mean(values))
    if feat & MEDIAN_FEATS:
        med = values[len(values) // 2]
        assert isinstance(med, float)
        return float(med)
    if feat & MODE_FEATS:
        return float(Counter(np.rint(values)).most_common(1)[0][0])
    if feat & KURTOSIS_FEATS:
        return float(_kurtosis(values))
    if feat & AUTO_CC_FEATS:
        mean = np.mean(values)
        c_1 = _acc_k(values, 1, mean)
        c_0 = _acc_k(values, 0, mean)
        return float(c_1 / c_0)
    if feat & MIN_FEATS:
        return float(np.min(values))
    if feat & MAX_FEATS:
        return float(np.max(values))
    if feat & RANGE_FEATS:
        return float(np.max(values) - np.min(values))
    if feat & STD_FEATS:
        return float(np.std(values))
    if feat & VAR_FEATS:
        return float(np.var(values))
    if feat & COEF_VAR_FEATS:
        mean = np.mean(values)
        return float(np.std(values) / mean if mean != 0 else 0)
    if feat & IQR_FEATS:
        val = np.percentile(values, 75) - np.percentile(values, 25)
        assert isinstance(val, float)
        return float(val)

    raise ValueError("Feature not found")


def get_feat_vector(traj: Trajectory, feats: int, **kwargs) -> np.ndarray:
    """Get feature vector from trajectory.

    Parameters
    ----------
    traj : Trajectory
        Trajectory.
    feats : int
        Features.

    Returns
    -------
    np.ndarray
        Feature vector.
    """
    return np.array(
        [get_feat(traj, feat, **kwargs) for feat in FEAT_VALUES if feat & feats]
    )



from pactus.features.features import (
    ACC_AUTO_CC,
    ACC_CHANGE_RATE_AUTO_CC,
    ACC_CHANGE_RATE_COEF_VAR,
    ACC_CHANGE_RATE_FEATS,
    ACC_CHANGE_RATE_IQR,
    ACC_CHANGE_RATE_KURTOSIS,
    ACC_CHANGE_RATE_MAX,
    ACC_CHANGE_RATE_MEAN,
    ACC_CHANGE_RATE_MEDIAN,
    ACC_CHANGE_RATE_MIN,
    ACC_CHANGE_RATE_MODE,
    ACC_CHANGE_RATE_RANGE,
    ACC_CHANGE_RATE_STD,
    ACC_CHANGE_RATE_VAR,
    ACC_COEF_VAR,
    ACC_FEATS,
    ACC_IQR,
    ACC_KURTOSIS,
    ACC_MAX,
    ACC_MEAN,
    ACC_MEDIAN,
    ACC_MIN,
    ACC_MODE,
    ACC_RANGE,
    ACC_STD,
    ACC_VAR,
    ALL,
    ANGLE_AUTO_CC,
    ANGLE_COEF_VAR,
    ANGLE_FEATS,
    ANGLE_IQR,
    ANGLE_KURTOSIS,
    ANGLE_MAX,
    ANGLE_MEAN,
    ANGLE_MEDIAN,
    ANGLE_MIN,
    ANGLE_MODE,
    ANGLE_RANGE,
    ANGLE_STD,
    ANGLE_VAR,
    AUTO_CC_FEATS,
    COEF_VAR_FEATS,
    DISTANCE,
    FEAT_NAMES,
    FEAT_VALUES,
    IQR_FEATS,
    KURTOSIS_FEATS,
    MAX_FEATS,
    MEAN_FEATS,
    MEDIAN_FEATS,
    MIN_FEATS,
    MODE_FEATS,
    RANGE_FEATS,
    STD_FEATS,
    STOP_RATE,
    TURNING_ANGLE_AUTO_CC,
    TURNING_ANGLE_CHANGE_RATE_AUTO_CC,
    TURNING_ANGLE_CHANGE_RATE_COEF_VAR,
    TURNING_ANGLE_CHANGE_RATE_FEATS,
    TURNING_ANGLE_CHANGE_RATE_IQR,
    TURNING_ANGLE_CHANGE_RATE_KURTOSIS,
    TURNING_ANGLE_CHANGE_RATE_MAX,
    TURNING_ANGLE_CHANGE_RATE_MEAN,
    TURNING_ANGLE_CHANGE_RATE_MEDIAN,
    TURNING_ANGLE_CHANGE_RATE_MIN,
    TURNING_ANGLE_CHANGE_RATE_MODE,
    TURNING_ANGLE_CHANGE_RATE_RANGE,
    TURNING_ANGLE_CHANGE_RATE_STD,
    TURNING_ANGLE_CHANGE_RATE_VAR,
    TURNING_ANGLE_COEF_VAR,
    TURNING_ANGLE_FEATS,
    TURNING_ANGLE_IQR,
    TURNING_ANGLE_KURTOSIS,
    TURNING_ANGLE_MAX,
    TURNING_ANGLE_MEAN,
    TURNING_ANGLE_MEDIAN,
    TURNING_ANGLE_MIN,
    TURNING_ANGLE_MODE,
    TURNING_ANGLE_RANGE,
    TURNING_ANGLE_STD,
    TURNING_ANGLE_VAR,
    VAR_FEATS,
    VEL_AUTO_CC,
    VEL_CHANGE_RATE,
    VEL_COEF_VAR,
    VEL_FEATS,
    VEL_IQR,
    VEL_KURTOSIS,
    VEL_MAX,
    VEL_MEAN,
    VEL_MEDIAN,
    VEL_MIN,
    VEL_MODE,
    VEL_RANGE,
    VEL_STD,
    VEL_VAR,
)
from pactus.features.featurizer import Featurizer



from __future__ import annotations

import json
import logging
from abc import ABCMeta, abstractmethod
from collections import Counter
from pathlib import Path
from typing import Any, Callable, List, Tuple, Union

from sklearn.model_selection import train_test_split
from yupi import Trajectory
from yupi.core import JSONSerializer

from pactus import config
from pactus.dataset._utils import _get_path, _get_progress_log


class Data:
    """
    Structure that groups the trajectories and labels along with some
    useful methods to work with the set of them.
    """

    def __init__(
        self, dataset: Dataset, trajs: List[Trajectory], labels: List[Any]
    ) -> None:
        self.dataset = dataset
        self.trajs = trajs
        self.labels = labels
        self.label_counts = Counter(labels)

    @property
    def classes(self) -> List[Any]:
        """Classes present in the dataset."""
        return list(self.label_counts.keys())

    def __len__(self) -> int:
        return len(self.trajs)

    def take(
        self,
        size: Union[float, int],
        stratify: bool = True,
        shuffle: bool = True,
        random_state: Union[int, None] = None,
    ) -> Data:
        """Takes a subset of the dataset."""
        if isinstance(size, int):
            assert 0 < size < len(self), "size should be within 0 and len(self)"
            size /= len(self)

        ans, _ = self.split(
            size, stratify=stratify, shuffle=shuffle, random_state=random_state
        )
        return ans

    def cut(self, size: Union[float, int]):
        """
        Similar to split, but without shuffle, stratify, etc. Just slices the
        dataset into two parts.
        """
        if isinstance(size, float):
            assert 0 < size < 1, "size should be within 0 and 1 if float"
            size = int(len(self) * size)
        else:
            assert (
                0 < size < len(self)
            ), "size should be within 0 and the dataset size if int"

        left, right = self.trajs[:size], self.trajs[size:]
        left_labels, right_labels = self.labels[:size], self.labels[size:]
        left_d = Data(self.dataset, left, left_labels)
        right_d = Data(self.dataset, right, right_labels)
        return left_d, right_d

    def split(
        self,
        train_size: Union[float, int] = 0.8,
        stratify: bool = True,
        shuffle: bool = True,
        random_state: Union[int, None] = None,
    ) -> Tuple[Data, Data]:
        """
        Splits the dataset into train and test dataset slices.

        It uses the sklearn.model_selection.train_test_split function.

        Parameters
        ----------
        train_size : Union[float, int], optional
            The proportion of the dataset to include in the train split.
            If float, should be between 0.0 and 1.0, if int, represents the
            absolute number of train samples. By default 0.8.
        stratify : bool, optional
            If True, the split will be stratified according to the labels,
            by default True
        shuffle : bool, optional
            If True, the split will be shuffled, by default True
        random_state : Union[int, None], optional
            Random seed for reproducibility, by default None

        Returns
        -------
        Tuple[Data, Data]
            A tuple with the train and test Data objects.
        """
        if isinstance(train_size, int):
            assert (
                0 < train_size < len(self)
            ), "train_size should be within 0 and the dataset size if int"
            train_size /= len(self)
        else:
            assert 0 < train_size < 1, "train_size should be within 0 and 1 if float"

        x_train, x_test, y_train, y_test = train_test_split(
            self.trajs,
            self.labels,
            train_size=train_size,
            stratify=self.labels if stratify else None,
            random_state=random_state,
            shuffle=shuffle,
        )

        train_data = Data(self.dataset, x_train, y_train)
        test_data = Data(self.dataset, x_test, y_test)
        return train_data, test_data

    def map(self, func: Callable[[Trajectory, Any], Tuple[Trajectory, Any]]) -> Data:
        """
        Applies a function to each trajectory and label pair.

        Usefull to apply some preprocessing to the trajectories
        or the labels.

        Parameters
        ----------
        func : Callable[[Trajectory, Any], Tuple[Trajectory, Any]]
            Function to be applied to each trajectory and label pair.

        Returns
        -------
        Data
            A new Data object with the results of the function.
        """
        trajs, labels = [], []
        for traj, label in zip(self.trajs, self.labels):
            traj, label = func(traj, label)
            trajs.append(traj)
            labels.append(label)
        return Data(self.dataset, trajs, labels)

    def filter(self, func: Callable[[Trajectory, Any], bool]) -> Data:
        """
        Filters the dataset based on a function.

        Parameters
        ----------
        func : Callable[[Trajectory, Any], bool]
            Function to be applied to each trajectory and label pair.

        Returns
        -------
        Data
            A new Data object with the filtered trajectories and labels.
        """
        trajs, labels = [], []
        for traj, label in zip(self.trajs, self.labels):
            if func(traj, label):
                trajs.append(traj)
                labels.append(label)
        logging.info("Filtered %d of %d trajectories", len(trajs), len(self))
        return Data(self.dataset, trajs, labels)


class Dataset(Data, metaclass=ABCMeta):
    """Wraps the data with some general properties that describes a full dataset"""

    def __init__(
        self, name: str, version: str, refetch: bool = False, reyupify: bool = False
    ):
        self.name = name
        self.version = version
        self.refetch = refetch
        self.reyupify = reyupify
        self._metadata = self._load_metadata()
        self._yupi_data: dict
        self.dir = _get_path(config.DS_DIR, self.name)
        self.raw_dir = _get_path(config.DS_RAW_DIR, self.name)

        trajs, labels = self.load()
        super().__init__(self, trajs, labels)

    def fetch(self) -> None:
        """Downloads the dataset in case needed"""

    @abstractmethod
    def yupify(self) -> Tuple[List[Trajectory], List[Any]]:
        """Parses or generates the dataset and convert it to yupi trajectories"""

    def __len__(self):
        return len(self.trajs)

    @property
    def metadata(self) -> dict:
        """Dataset metadata"""
        assert (
            self._metadata is not None
        ), f"There is not metadata loaded for the {self.name} dataset."
        return self._metadata

    @property
    def _has_metadata(self) -> bool:
        return self._metadata is not None

    def _default_metadata(self) -> dict:
        metadata = {
            "name": self.name,
            "path": str(_get_path(config.DS_METADATA_FILE, self.name)),
            "version": self.version,
            "yupi_data": None,
        }
        return metadata

    def _load_metadata(self) -> Union[dict, None]:
        metadata_path = _get_path(config.DS_METADATA_FILE, self.name)
        if not metadata_path.exists():
            return None
        with open(metadata_path, "r", encoding="utf-8") as metadata_file:
            return json.load(metadata_file)

    def _refetch_required(self) -> bool:
        """Checks if the current cache requires a refetch."""
        needs_refetch = self.refetch or not self._has_metadata

        if self._has_metadata:
            stored_major = self.metadata["version"].split(".")[0]
            current_major = self.version.split(".")[0]
            needs_refetch = needs_refetch or stored_major != current_major

        return needs_refetch

    def _yupify_required(self) -> bool:
        """Checks if the current cache requires a reyupify."""
        needs_yupify = self.reyupify or not self._has_metadata

        if self._has_metadata:
            stored_minor = self.metadata["version"].split(".")[1]
            current_minor = self.version.split(".")[1]
            yupi_data_file = self.dir / "yupi_data.json"
            needs_yupify = (
                needs_yupify
                or stored_minor != current_minor
                or not yupi_data_file.exists()
            )

        if needs_yupify:
            logging.info("Yupify is required for the %s dataset", self.name)
        return needs_yupify

    def _save_json(self, path: Path, data: dict):
        with open(path, "w", encoding="utf-8") as md_file:
            json.dump(data, md_file, ensure_ascii=False, indent=4)

    def _update_metadata(self):
        logging.info("Updating metadata for %s dataset", self.name)
        metadata_path = Path(self.metadata["path"])
        self._save_json(metadata_path, self.metadata)

    def _create_folder(self):
        """Create dataset folder if not exists"""
        dataset_path = _get_path(config.DS_DIR, self.name)
        dataset_path.mkdir(parents=True, exist_ok=True)

    def _yupify(self):
        """
        Gets the trajectoires in a yupi format and the labels
        and stores them
        """
        logging.info("Yupifying %s dataset", self.name)
        trajs, labels = self.yupify()

        ds_dir = _get_path(config.DS_DIR, self.name)

        for i, traj in enumerate(trajs):
            traj.traj_id = str(i)

        json_trajs = [JSONSerializer.to_json(traj) for traj in trajs]
        yupi_data = {"trajs": json_trajs, "labels": labels}

        logging.info("Saving yupify trajectories for %s dataset", self.name)
        data_path = ds_dir / "yupi_data.json"
        self._save_json(data_path, yupi_data)
        self.metadata["yupi_data"] = str(data_path)

    def _ensure_cache(self):
        if self._refetch_required():
            logging.info("Fetching %s dataset", self.name)
            self.fetch()
            self._metadata = self._default_metadata()
            self._update_metadata()

        if self._yupify_required():
            self._yupify()

        self.metadata["version"] = self.version
        self._update_metadata()

    def _load_yupi_data(self):
        assert self.metadata["yupi_data"] is not None
        logging.info("Loading yupify data for %s dataset", self.name)
        yupi_metadata_path = self.metadata["yupi_data"]
        with open(yupi_metadata_path, "r", encoding="utf-8") as md_file:
            self._yupi_data = json.load(md_file)

    def _load(self) -> Tuple[List[Trajectory], List[Any]]:
        self._load_yupi_data()
        logging.info("Loading %s dataset", self.name)

        def _load_traj(traj, i, total):
            print(_get_progress_log(i, total), end="\r")
            return JSONSerializer.from_json(traj)

        total = len(self._yupi_data["trajs"])
        trajs = [
            _load_traj(traj, i + 1, total)
            for i, traj in enumerate(self._yupi_data["trajs"])
        ]
        labels = self._yupi_data["labels"]
        return trajs, labels

    def load(self) -> Tuple[List[Trajectory], List[Any]]:
        """Loads the dataset"""

        self._create_folder()
        self._ensure_cache()
        return self._load()



import logging
from pathlib import Path

import numpy as np
from yupi import Trajectory

from pactus import config
from pactus.dataset import Data
from pactus.dataset._utils import _get_path
from pactus.features.features import get_feat_vector

FeatureMask = int


class Featurizer:
    """
    The featurizer class is used to compute selected features of a given
    trajectory or dataset.
    """

    def __init__(self, selected: FeatureMask, recompute: bool = False, **kwargs):
        self.selected = selected
        self.recompute = recompute
        self.kwargs = kwargs

        self.kwargs["stop_rate_threshold"] = self.kwargs.get("stop_rate_threshold", 1)
        self.kwargs["vel_change_rate_threshold"] = self.kwargs.get(
            "vel_change_rate_threshold", 1
        )

    def _recompute_required(self, feat_file: Path) -> bool:
        return self.recompute or not feat_file.exists()

    def _recompute(
        self,
        data: Data,
        feat_file: Path,
    ) -> np.ndarray:
        logging.info("Recomputing features")
        feats = np.stack(
            [get_feat_vector(traj, self.selected, **self.kwargs) for traj in data.trajs]
        )
        np.savetxt(str(feat_file), feats)
        return feats

    def compute(self, data: Data) -> np.ndarray:
        """Computes the features matrix for a given dataset or slice."""

        feats = None
        dataset = data.dataset
        trajs = data.trajs
        file_name = f"{self.selected}.txt"
        feat_file = _get_path(config.DS_FEATS_DIR, dataset.name) / file_name
        feat_file.parent.mkdir(parents=True, exist_ok=True)

        # recompute if required
        if self._recompute_required(feat_file):
            feats = self._recompute(dataset, feat_file)

        if feats is None:
            logging.info("Loading features from file")
            feats = np.loadtxt(feat_file)

        idx = np.array([int(traj.traj_id or "") for traj in trajs])
        trajs_feats = feats[idx]
        return trajs_feats

    def traj2vec(self, traj: Trajectory) -> np.ndarray:
        """Computes the features vector for a given trajectory."""
        return get_feat_vector(traj, self.selected, **self.kwargs)
